package main

import (
	"bytes"
	"converge/pkg/converge"
	"converge/pkg/websocketutil"
	"fmt"
	"io"
	"log"
	"net"
	"net/http"
	"os"
	"regexp"
	"strings"
	"sync"
)

func parsePublicId(path string) (publicId string, _ error) {
	pattern := regexp.MustCompile("^/[^/]+/([^/]+)$")
	matches := pattern.FindStringSubmatch(path)
	if len(matches) != 2 {
		return "", fmt.Errorf("Invalid URL path '%s'", path)
	}
	return matches[1], nil
}

func catchAllHandler(w http.ResponseWriter, r *http.Request) {
	http.Redirect(w, r, "/docs", http.StatusFound)
	return
}

// filters to filter html content.
var filters map[string]string = make(map[string]string)

type FileHandlerFilter struct {
	http.Handler
}

func (handler FileHandlerFilter) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	mutex := sync.Mutex{}
	// pedantic
	mutex.Lock()

	if r.TLS == nil {
		filters["secure"] = ""
	} else {
		filters["secure"] = "s"
	}
	for _, header := range []string{"X-Forwarded-Proto", "X-Scheme", "X-Forwarded-Scheme"} {
		values := r.Header.Values(header)
		for _, value := range values {
			if strings.ToLower(value) == "https" {
				filters["secure"] = "s"
			}
		}
	}
	filters["host"] = r.Host
	mutex.Unlock()

	w.Header().Set("Cache-Control", "no-cache, no-store, must-revalidate")
	w.Header().Set("Pragma", "no-cache")
	w.Header().Set("Expires", "0")

	handler.Handler.ServeHTTP(w, r)
}

type FilteredFileSystem struct {
	fs http.FileSystem
}

func (ffs FilteredFileSystem) Open(name string) (http.File, error) {
	f, err := ffs.fs.Open(name)
	if err != nil {
		return nil, err
	}
	if !strings.HasSuffix(name, ".html") {
		return f, nil
	}
	return NewFilteredFile(f, filters)
}

type FilteredFile struct {
	http.File

	// Read bytes 0..(index-1) already
	index        int
	fullContents *bytes.Buffer
}

func NewFilteredFile(f http.File, filter map[string]string) (FilteredFile, error) {
	file := FilteredFile{
		index:        0,
		fullContents: bytes.NewBuffer(make([]byte, 0)),
	}
	file.File = f

	_, err := io.Copy(file.fullContents, file.File)
	if err != nil {
		return FilteredFile{}, err
	}
	contents := file.fullContents.String()
	for key, value := range filter {
		key = "@" + key + "@"
		contents = strings.ReplaceAll(contents, key, value)
	}
	file.fullContents = bytes.NewBufferString(contents)

	return file, nil
}

func (ff FilteredFile) Read(p []byte) (n int, err error) {
	return ff.fullContents.Read(p)
}

func main() {

	downloadDir := "downloads"
	if len(os.Args) == 2 {
		downloadDir = os.Args[1]
	}
	log.Println("Doanload directory", downloadDir)

	admin := converge.NewAdmin()
	registrationService := websocketutil.WebSocketService{
		Handler: func(w http.ResponseWriter, r *http.Request, conn net.Conn) {
			publicId, err := parsePublicId(r.URL.Path)
			if err != nil {
				log.Printf("Cannot parse public id from url: '%v'\n", err)
				return
			}
			log.Printf("Got registration connection: '%s'\n", publicId)
			err = admin.Register(publicId, conn)
			if err != nil {
				log.Printf("Error %v\n", err)
			}
		},
	}
	clientService := websocketutil.WebSocketService{
		Handler: func(w http.ResponseWriter, r *http.Request, conn net.Conn) {
			publicId, err := parsePublicId(r.URL.Path)
			if err != nil {
				log.Printf("Cannot parse public id from url: '%v'\n", err)
				return
			}
			log.Printf("Got client connection: '%s'\n", publicId)
			err = admin.Connect(publicId, conn)
			if err != nil {
				log.Printf("Error %v\n", err)
			}
		},
	}

	http.HandleFunc("/agent/", registrationService.Handle)
	http.HandleFunc("/client/", clientService.Handle)

	filesystem := http.Dir(downloadDir)
	filteredFilesystem := FilteredFileSystem{
		fs: filesystem,
	}
	fileHandler := FileHandlerFilter{http.FileServer(filteredFilesystem)}

	http.Handle("/docs/", http.StripPrefix("/docs/", fileHandler))
	http.HandleFunc("/", catchAllHandler)

	// Start HTTP server
	fmt.Println("Rendez-vous server listening on :8000")
	log.Fatal(http.ListenAndServe(":8000", nil))
}
